/***************************************************************************
 *   Copyright (C) 2006 by Michael Andreen                                 *
 *   andreen@student.chalmers.se                                           *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA          *
 ***************************************************************************/

#include "genealgorithms.h"
#include "geneorder.h"

#include <algorithm>
#include <set>
#include <stack>
#include <cstdlib>
#include <iostream>
using namespace std;

std::pair<int,int> longestSequences(const GeneOrder& go){
	vector<vector<int> > v = robinsonSchensted(go);
	return pair<int,int>(v[0].size(),v.size());
}

std::vector<std::vector<int> > robinsonSchensted(const GeneOrder& go){
	vector<vector<int> > v;
	for (GeneOrder::iterator i = go.begin(); i != go.end(); ++i){
		int n = abs(*i);
		bool added = false;
		for (vector<vector<int> >::iterator vs = v.begin();
				vs != v.end(); ++vs){
			vector<int>::iterator bigger = upper_bound(vs->begin(),vs->end(),n);
			if ( bigger == vs->end()){
				vs->push_back(n);
				added = true;
				break;
			}else{
				swap(n,*bigger);
			}
		}
		if (!added){
			v.push_back(vector<int>());
			v.back().push_back(n);
		}
	}
	return v;
}

struct FindP{
	size_t p;
	FindP(size_t p) : p(p) {}
	bool operator()(Interval i){
		return (i.first == p || i.second == p);
	}
};


int countCycles(const GeneOrder& go){
	int cycles = 0;
	set<size_t> marked;
	vector<Interval> intervals = findIntervals(go);
	vector<Interval> points = findIntervalsAtPoints(intervals);
	for (size_t p = 1; p < go.size(); ++p){
		if (marked.find(p) != marked.end())
			continue;
		Interval i = intervals[points[p].first];
		while (marked.find(p) == marked.end()){
			marked.insert(p);
			if (i == intervals[points[p].first])
				i = intervals[points[p].second];
			else
				i = intervals[points[p].first];

			if (p == i.first)
				p = i.second;
			else
				p = i.first;
		}
		++cycles;
	}
	return cycles;
}

int sign(Gene g){
	if (g > 0)
		return 1;
	if (g < 0)
		return -1;
	return 0;
}

struct Abs{
	Gene operator()(Gene x) const{
		return abs(x);
	}
};
std::vector<Component> findComponents(const GeneOrder& go){
	vector<Component> components;
	vector<int> os(go.size()-1);
	for (size_t i = 0; i < os.size(); ++i)
		os[i] = (go[i]*go[i+1] > 0 ? sign(go[i]) : 0);
	stack<Gene> Mdir;
	Mdir.push(go.size()-1);
	stack<Gene> Mrev;
	Mrev.push(0);
	stack<size_t> Sdir;
	Sdir.push(0);
	stack<size_t> Srev;
	Srev.push(0);
	vector<Gene> dir;
	dir.push_back(go.size()-1);
	vector<Gene> rev;
	rev.push_back(0);
	size_t s;
	vector<Gene> p(go.list());
	transform(p.begin(),p.end(),p.begin(),Abs());
	for (size_t i = 1; i < go.size(); ++i){
		//Directed
		if (p[i-1] > p[i])
			Mdir.push(p[i-1]);
		else while (Mdir.top() < p[i])
			Mdir.pop();
		dir.push_back(Mdir.top());

		s = Sdir.top();
		while(p[Sdir.top()] > p[i] || dir[Sdir.top()] < p[i]){
			Sdir.pop();
			os[Sdir.top()] = (os[Sdir.top()] == os[s] ? os[s] : 0);
			s = Sdir.top();
		}
		if (go[i] > 0 && dir[i] == dir[s] && static_cast<Gene>(i - s) == p[i] - p[s])
			components.push_back(Component(p[s],p[i],(s+1 == i ? 0 : os[s])));

		//Reverse
		if (p[i-1] < p[i])
			Mrev.push(p[i-1]);
		else while (Mrev.top() > p[i])
			Mrev.pop();
		rev.push_back(Mrev.top());

		s = Srev.top();
		while((p[s] < p[i] || rev[s] > p[i]) && s > 0){
			Srev.pop();
			os[Srev.top()] *= (os[Srev.top()] == os[s] ? 1 : 0);
			s = Srev.top();
		}
		if (go[i] < 0 && rev[i] == rev[s] && static_cast<Gene>(i - s) == p[s] - p[i])
			components.push_back(Component(-p[s],-p[i],(s+1 == i ? 0 : os[s])));

		//Update stacks
		if (go[i] > 0)
			Sdir.push(i);
		else
			Srev.push(i);
	}
	return components;
}

/**
 * 
 */
std::vector<Interval> findIntervals(const GeneOrder& go){
	vector<Interval> intervals(go.size()-1,Interval(go.size(),go.size()));
	size_t n = 0;
	for (GeneOrder::iterator g = go.begin(); g != go.end(); ++g, ++n){
			size_t i = abs(*g);
			if (i < go.size() - 1)
				intervals[i].first = n + (*g >= 0 ? 1 : 0);
			if (i > 0)
				intervals[i-1].second = n + (*g < 0 ? 1 : 0);
	}
	return intervals;
}

/**
 * 
 */
std::vector<Interval> findIntervalsAtPoints(const vector<Interval>& intervals){
	size_t max = intervals.size()+1;
	vector<Interval> points(max,Interval(max,max));
	size_t n = 0;
	for (vector<Interval>::const_iterator i = intervals.begin(); i != intervals.end(); ++i, ++n){
		if (points[i->first].first == max){
			points[i->first].first = n;
		}else
			points[i->first].second = n;
			
		if (points[i->second].first == max){
			points[i->second].first = n;
		}else
			points[i->second].second = n;
	}
	return points;

}
